/********************************************************************\
 * Mu (Miscellaneous Utilities) is a general convenience library.   *
 * Copyright (C) 2020  Asher Gordon <AsDaGo@posteo.net>             *
 *                                                                  *
 * This file is part of Mu.                                         *
 *                                                                  *
 * Mu is free software: you can redistribute it and/or modify it    *
 * under the terms of the GNU General Public License as published   *
 * by the Free Software Foundation, either version 3 of the         *
 * License, or (at your option) any later version.                  *
 *                                                                  *
 * Mu is distributed in the hope that it will be useful, but        *
 * WITHOUT ANY WARRANTY; without even the implied warranty of       *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the    *
 * GNU General Public License for more details.                     *
 *                                                                  *
 * You should have received a copy of the GNU General Public        *
 * License along with Mu.  If not, see                              *
 * <https://www.gnu.org/licenses/>.                                 *
\********************************************************************/

/*
 * safe.h -- versions of standard functions that do not return on
 * error
 */

#ifndef _MU_SAFE_H
#define _MU_SAFE_H

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>

#include "compat.h"
#include "format.h"
#include "options.h"

/* For use in __attribute__(). Offset by 3 for `file', `line',
   `func'. */
#define _mu_safe_format_attr(archetype, fmt_ind, first)		\
  format(archetype, (fmt_ind) + 3, (first) ? (first) + 3 : 0)

#define _MU_SAFE_TR_NAME(name) _##name##_line
#define _MU_SAFE_DECL(name, ...)					\
  _MU_SAFE_TR_NAME(name)						\
    (const char *file, int line, const char *func, __VA_ARGS__)
#define _MU_SAFE_DEF(name, ...)						\
  _MU_SAFE_TR_NAME(name)(__FILE__, __LINE__, __func__, __VA_ARGS__)

/* Format is part of the '...' so there is always at least one
   __VA_ARGS__ for a valid call of mu_die(). */
#define mu_die(stat, ...)	_MU_SAFE_DEF(mu_die, stat, __VA_ARGS__)
#define mu_warn(...)		_MU_SAFE_DEF(mu_warn, __VA_ARGS__)
#define mu_vdie(stat, fmt, ap)	_MU_SAFE_DEF(mu_vdie, stat, fmt, ap)
#define mu_vwarn(fmt, ap)	_MU_SAFE_DEF(mu_vwarn, fmt, ap)

#define mu_xmalloc(size)	_MU_SAFE_DEF(mu_xmalloc, size)
#define mu_xcalloc(nmemb, size)	_MU_SAFE_DEF(mu_xcalloc, nmemb, size)
#define mu_xrealloc(ptr, size)	_MU_SAFE_DEF(mu_xrealloc, ptr, size)
#define mu_xreallocarray(ptr, nmemb, size)		\
  _MU_SAFE_DEF(mu_xreallocarray, ptr, nmemb, size)
#define mu_xstrdup(str)		_MU_SAFE_DEF(mu_xstrdup, str)
#define mu_xasprintf(strp, ...)	_MU_SAFE_DEF(mu_xasprintf, strp, __VA_ARGS__)
#define mu_xvasprintf(strp, fmt, ap)		\
  _MU_SAFE_DEF(mu_xvasprintf, strp, fmt, ap)
#define mu_xformat(s, w, c, i, si, ...)			\
  _MU_SAFE_DEF(mu_xformat, s, w, c, i, si, __VA_ARGS__)
#define mu_xformat_string(w, c, i, s, ...)		\
  _MU_SAFE_DEF(mu_xformat, w, c, i, s, __VA_ARGS__)
#define mu_xvformat(s, w, c, i, si, f, a)		\
  _MU_SAFE_DEF(mu_xvformat, s, w, c, i, si, f, a)
#define mu_xvformat_string(w, c, i, s, f, a)		\
  _MU_SAFE_DEF(mu_xvformat_string, w, c, i, s, f, a)
#define mu_xformat_help(s, n, u, d, nt, o)		\
  _MU_SAFE_DEF(mu_xformat_help, s, n, u, d, nt, o)
#define mu_xformat_help_string(n, u, d, nt, o, g, w)		\
  _MU_SAFE_DEF(mu_xformat_help_string, n, u, d, nt, o, g, w)
#define mu_xformat_help_man(s, n, u, sd, d, nt, sc, sn, p, dt, o)\
  _MU_SAFE_DEF(mu_xformat_help_man(s, n, u, sd, d, nt,		 \
				   sc, sn, p, dt, o))
#define mu_xformat_help_man_string(n, u, sd, d, nt, sc, sn, p, dt, o)\
  _MU_SAFE_DEF(mu_xformat_help_man_string(n, u, sd, d, nt,	     \
					  sc, sn, p, dt, o))
#define mu_xconcat_opts(res, a, b, extra)		\
  _MU_SAFE_DEF(mu_xconcat_opts, res, a, b, extra)

void _MU_SAFE_DECL(mu_die, int, const char *, ...)
  __attribute__((_mu_safe_format_attr(printf, 2, 3), noreturn));
int _MU_SAFE_DECL(mu_warn, const char *, ...)
  __attribute__((_mu_safe_format_attr(printf, 1, 2)));
void _MU_SAFE_DECL(mu_vdie, int, const char *, va_list)
  __attribute__((_mu_safe_format_attr(printf, 2, 0), noreturn));
int _MU_SAFE_DECL(mu_vwarn, const char *, va_list)
  __attribute__((_mu_safe_format_attr(printf, 1, 0)));

void *_MU_SAFE_DECL(mu_xmalloc, size_t) __attribute__((malloc));
void *_MU_SAFE_DECL(mu_xcalloc, size_t, size_t)
  __attribute__((malloc));
void *_MU_SAFE_DECL(mu_xrealloc, void *, size_t);
void *_MU_SAFE_DECL(mu_xreallocarray, void *, size_t, size_t);
char *_MU_SAFE_DECL(mu_xstrdup, const char *) __attribute__((malloc));
unsigned _MU_SAFE_DECL(mu_xasprintf, char **, const char *, ...)
  __attribute__((_mu_safe_format_attr(printf, 2, 3)));
unsigned _MU_SAFE_DECL(mu_xvasprintf, char **, const char *, va_list)
  __attribute__((_mu_safe_format_attr(printf, 2, 0)));
void _MU_SAFE_DECL(mu_xformat,
		   FILE *, unsigned short, unsigned short, unsigned short *,
		   unsigned short, unsigned short, const char *, ...)
  __attribute__((_mu_safe_format_attr(printf, 7, 8)));
char *_MU_SAFE_DECL(mu_xformat_string,
		    unsigned short, unsigned short, unsigned short *,
		    unsigned short, unsigned short, const char *, ...)
  __attribute__((_mu_safe_format_attr(printf, 6, 7), malloc));
void _MU_SAFE_DECL(mu_xvformat,
		   FILE *, unsigned short, unsigned short, unsigned short *,
		   unsigned short, unsigned short, const char *, va_list)
  __attribute__((_mu_safe_format_attr(printf, 7, 0)));
char *_MU_SAFE_DECL(mu_xvformat_string,
		    unsigned short, unsigned short, unsigned short *,
		    unsigned short, unsigned short, const char *, va_list)
  __attribute__((_mu_safe_format_attr(printf, 6, 0), malloc));
void _MU_SAFE_DECL(mu_xformat_help,
		   FILE *, const char *, const char *,
		   const char *, const char *, const MU_OPT *);
char *_MU_SAFE_DECL(mu_xformat_help_string,
		    const char *, const char *, const char *,
		    const char *, const MU_OPT *,
		    unsigned short, unsigned short)
  __attribute__((malloc));
void _MU_SAFE_DECL(mu_xformat_help_man,
		   FILE *, const char *, const char *, const char *,
		   const char *, const char *, const char *,
		   const char *, const char *, const char *,
		   const MU_OPT *);
char *_MU_SAFE_DECL(mu_xformat_help_man_string,
		    const char *, const char *, const char *,
		    const char *, const char *, const char *,
		    const char *, const char *, const char *,
		    const MU_OPT *)
  __attribute__((malloc));
size_t _MU_SAFE_DECL(mu_xconcat_opts,
		     MU_OPT **, const MU_OPT *, const MU_OPT *, size_t);

#endif /* !_MU_SAFE_H */
